# -*- coding: Latin-1 -*-
#
# SMM Share and Money Management
# Copyright (c) 2007-2010 Carsten Grohmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

## \file gui/log.py
#
# SMM specific functions and classes to record and print out log messages
#
# \namespace gui::log
#
# This module provides a own logger class as well as specific funtions to
# improve Pythons logging facility.
#
# The SMM GUI uses the python logging instance with three %log handler
# attached.
#
# The first handler (StringHandler) is used to cache messages later
# displaying calling getBufferAsList() or getBufferAsString().
#
# The second handler (StreamHandler) to print error messages to
# \c sys.stderr.
#
# The third handler (FileHandler) writes all messages into a file. This
# behaviour is useful to store logged expections permanently.
#
# \note The classes StreamHandler and FileHandler are copiied from the current
# revision (76551 - 27.11.2009) of Python's public SVN repository. Because the
# implementations included in Python 2.5 and some versions of Python 2.6
# aren't ready to handle Unicode properly.
#
# \author Carsten Grohmann <mail (at) carstengrohmann (dot) de>
#
# \par License:
# GPL version 2.0

import codecs
import logging
from logging import Handler
import logging.handlers
import os
import pprint
import sys
import traceback
import types


## \var stringLoggerInstance
#  Reference to the active StringHandler instance
stringLoggerInstance = None


## Contains the original exception handler
#
#  \see installExceptionHandler() and deinstallExceptionHandler()
_orig_exec_handler = None


## Stores the %log records as a list of strings.
class StringHandler(logging.handlers.MemoryHandler):

    ## \var storeAsUnicode
    #  Stores the %log records as unicode strings

    ## \var buffer
    #  The message buffer itself

    ## Encoding of all character strings
    #
    #  The \c default_encoding is used to convert character strings
    #  into unicode strings
    #
    #  \see storeAsUnicode
    encoding = sys.stdout.encoding or sys.getfilesystemencoding()

    ## Constructor
    #
    #  \param self           The object instance itself
    #  \param storeAsUnicode Store recorded %log records as unicode strings
    def __init__(self, storeAsUnicode=True):
        logging.handlers.MemoryHandler.__init__(self, sys.maxint, 99)
        self.storeAsUnicode = storeAsUnicode

    ## Convert a non unicode string into a unicode string
    #
    #  \return The unicode string
    #  \see encoding
    def _toUnicode(self, msg):
        # return msg if is already a unicode string or None
        if type(msg) in [types.UnicodeType, types.NoneType]:
            return msg

        # convert character string into a unicode string
        if type(msg) != types.UnicodeType:
            msg = msg.decode(self.encoding, 'replace')
        return msg

    ## Returns all buffered messages
    #
    #  \param clean Clean the internal message buffer
    #  \return List of messages strings
    #
    #  \see getBufferAsString()
    def getBufferAsList(self, clean=True):
        self.acquire()
        try:
            messages = self.buffer[:]
            if clean:
                self.flush()
        finally:
            self.release()
        return messages

    ## Returns all buffered messages
    #
    #  \param clean Clean the internal message buffer
    #  \return Messages concatenate to a string
    #
    #  \see getBufferAsList(self,)
    def getBufferAsString(self, clean=True):
        msg_list = getBufferAsList(clean)
        if self.storeAsUnicode:
            return u'\n'.join(msg_list)
        return '\n'.join(msg_list)

    ## Emit a record.
    #
    #  Add a formatted log record to the buffer.
    def emit(self, record):
        msg = self.format(record)
        if self.storeAsUnicode:
            msg = self._toUnicode(msg)
        self.buffer.append(msg)
        if self.shouldFlush(record):
            self.flush()

    ## Empty the buffer
    def flush(self):
        self.buffer = []


## Initialise the logging facility
#
#  Initialise and configure the logging itself as well as the handlers
#  described above. 
#
#  Our own execption handler will be installed finally.
#
#  \param filename Name of the log file
#  \param encoding Encoding of the log file
#  \param level    Verbosity of messages written in log file
#
#  \see StringHandler, stringLoggerInstance and installExceptionHandler()
def init(filename='smm.log', encoding=None, level=None):
    global stringLoggerInstance

    default_formatter = logging.Formatter('%(levelname)-8s: %(message)s')
    file_formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s: %(message)s')
    logger = logging.getLogger()

    console = StreamHandler()
    console.setLevel(logging.ERROR)
    console.setFormatter(default_formatter)

    fileLogger = FileHandler(filename, encoding=encoding)
    fileLogger.setFormatter(file_formatter)
    fileLogger.setLevel(logging.NOTSET)

    stringLoggerInstance = StringHandler()
    stringLoggerInstance.setLevel(logging.INFO)
    stringLoggerInstance.setFormatter(default_formatter)

    # check for installed handlers and remove them
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    # add new handler
    logger.addHandler(stringLoggerInstance)
    logger.addHandler(console)
    logger.addHandler(fileLogger)
    logger.setLevel(logging.NOTSET)

    # Set loglevel for file logger only
    if level:
        if level.upper() in logging._levelNames:
            fileLogger.setLevel(logging._levelNames[level.upper()])
        else:
            logging.warning(
                _(u'Ungltiges Log-Level "%s". Setze "WARNING"'),
                level.upper(),
                )
            fileLogger.setLevel(logging.WARNING)
    else:
        fileLogger.setLevel(logging.NOTSET)

    # inject own function to log exception details
    logging.exception = exception

    # install own exception handler
    installExceptionHandler()


## Reactivate system exception handler
#
#  \see deinstallExceptionHandler()
def deinit():
    deinstallExceptionHandler()


## \copydoc StringHandler::getBufferAsList()
def getBufferAsList(clean=True):
    return stringLoggerInstance.getBufferAsList(clean)


## \copydoc StringHandler::getBufferAsString()
def getBufferAsString(clean=True):
    return stringLoggerInstance.getBufferAsString(clean)


## \copydoc StringHandler::flush()
def flush():
    stringLoggerInstance.flush()


## Logs an exception message and the formatted traceback
#
#  \see exceptionHandler()
def exception(msg):
    logging.error(msg)
    if not sys.exc_info()[0]:
        return
    exceptionHandler(*sys.exc_info())
    sys.exc_clear()


## Install own exception handler
#
#  The original exception handler is saved in \ref _orig_exec_handler.
#
#  \see \ref _orig_exec_handler and deinstallExceptionHandler()
def installExceptionHandler():
    global _orig_exec_handler
    if _orig_exec_handler:
        logging.debug(
            _(u'Die Funktion zum Protokollieren von Programmfehlern ist bereits installiert.'),
            )
        return

    _orig_exec_handler = sys.excepthook
    sys.excepthook = exceptionHandler


## Restore the original exception handler
#
#  The original exception handler has been saved in \ref _orig_exec_handler.
#
#  \see \ref _orig_exec_handler and installExceptionHandler()
def deinstallExceptionHandler():
    if not _orig_exec_handler:
        logging.debug(
            _(u'Es ist keine eigene Funktion zum Protokollieren von Programmfehlern installiert. Daher kann nichts deinstalliert werden.'),
            )
        return

    sys.excepthook = _orig_exec_handler


## Logs detailed information about uncatched exceptions
#
#  \param exc_type  Type of the exception (normally a class object)
#  \param exc_value The "value" of the exception
#  \param exc_tb    Call stack of the exception
def exceptionHandler(exc_type, exc_value, exc_tb):
    try:
        tb = exc_tb
        while tb.tb_next:
            tb = tb.tb_next
        frame_locals = tb.tb_frame.f_locals

        # log exception details
        logging.error(_(u'Es trat ein ein unerwarteter Programmfehler auf.'))
        logging.error(_(u'Art des Programmfehlers: %s'), exc_type)
        logging.error(_(u'Fehlerdetails:           %s'), exc_value)
        logging.error(_(u'Stack Trace:'))
        lines = '\n'.join(traceback.format_exception(exc_type, exc_value, exc_tb)).split('\n')
        for line in lines:
            if not line:
                continue
            logging.error(line.decode('ascii', 'replace'))
        logging.error(_(u'Lokale Variablen des letzten Eintrags auf dem Stack:'))
        for var in frame_locals.keys():
            # convert variablen name and value to unicode
            uvar = var.decode('ascii', 'replace')
            uvalue = pprint.pformat(frame_locals[var]).decode('ascii', 'replace')

            logging.error(_(u'%s: %s'), uvar, uvalue)

    # delete local references of tracebacks or part of tracebacks
    # to avoid cirular references
    finally:
        del tb
        del frame_locals


class StreamHandler(Handler):
    """
    A handler class which writes logging records, appropriately formatted,
    to a stream. Note that this class does not close the stream, as
    sys.stdout or sys.stderr may be used.
    """

    def __init__(self, stream=None):
        """
        Initialize the handler.

        If stream is not specified, sys.stderr is used.
        """
        Handler.__init__(self)
        if stream is None:
            stream = sys.stderr
        self.stream = stream

    def flush(self):
        """
        Flushes the stream.
        """
        if self.stream and hasattr(self.stream, "flush"):
            self.stream.flush()

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            msg = self.format(record)
            stream = self.stream
            fs = "%s\n"
            try:
                if (isinstance(msg, unicode) and
                    getattr(stream, 'encoding', None)):
                    fs = fs.decode(stream.encoding)
                    try:
                        stream.write(fs % msg)
                    except UnicodeEncodeError:
                        #Printing to terminals sometimes fails. For example,
                        #with an encoding of 'cp1251', the above write will
                        #work if written to a stream opened or wrapped by
                        #the codecs module, but fail when writing to a
                        #terminal even when the codepage is set to cp1251.
                        #An extra encoding step seems to be needed.
                        stream.write((fs % msg).encode(stream.encoding))
                else:
                    stream.write(fs % msg)
            except UnicodeError:
                stream.write(fs % msg.encode("UTF-8"))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


class FileHandler(StreamHandler):
    """
    A handler class which writes formatted logging records to disk files.
    """
    def __init__(self, filename, mode='a', encoding=None, delay=0):
        """
        Open the specified file and use it as the stream for logging.
        """
        #keep the absolute path, otherwise derived classes which use this
        #may come a cropper when the current directory changes
        if codecs is None:
            encoding = None
        self.baseFilename = os.path.abspath(filename)
        self.mode = mode
        self.encoding = encoding
        if delay:
            #We don't open the stream, but we still need to call the
            #Handler constructor to set level, formatter, lock etc.
            Handler.__init__(self)
            self.stream = None
        else:
            StreamHandler.__init__(self, self._open())

    def close(self):
        """
        Closes the stream.
        """
        if self.stream:
            self.flush()
            if hasattr(self.stream, "close"):
                self.stream.close()
            StreamHandler.close(self)
            self.stream = None

    def _open(self):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        """
        if self.encoding is None:
            stream = open(self.baseFilename, self.mode)
        else:
            stream = codecs.open(self.baseFilename, self.mode, self.encoding)
        return stream

    def emit(self, record):
        """
        Emit a record.

        If the stream was not opened because 'delay' was specified in the
        constructor, open it before calling the superclass's emit.
        """
        if self.stream is None:
            self.stream = self._open()
        StreamHandler.emit(self, record)
