# -*- coding: Latin-1 -*-
#
# SMM Share and Money Management
# Copyright (c) 2007-2010 Carsten Grohmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

## \file smm/objects.py
#
# \short All data classes using STORM
#
# \author Carsten Grohmann <mail (at) carstengrohmann (dot) de>
#
# \par License:
# GPL version 2.0


# import general python modules
import datetime
import decimal
import logging
import types
import sys

# import project modules
from storm.locals import *

import database


## \brief Base class for all %database related classes
class SMMBaseObject(Storm):
    
    ## \brief Set of argument names for keyword arguments
    #
    #  This functions should contains a set of all possible arguments
    #  names which can by used as keyword arguments.
    #
    #  The primary key argument (mostly \c id) shouldn't be part of
    #  this set, because this argument will be setted by the Storm
    #  framework.
    #
    #  Attributes marked \c unicode or \c decimal will be converted to
    #  the specified type automatically.
    #
    #  \see _toUnicode() and _toDecimal()
    arguments = {
        'name': 'unicode',
        }


    ## \brief Encoding of all character strings
    #
    #  The \c default_encoding is used to convert character strings
    #  into unicode strings
    #  \note Use the system encoding not a hard coded encoding
    encoding = sys.stdout.encoding or sys.getfilesystemencoding()


    ## \brief Initialise the instance and set attributes optionally
    #
    #  This constructor creates a new object, without checking the store
    #  whether an object exists. Use the get* functions to check the
    #  store first and creates on object later.
    #
    #  Call \ref add2db() to store a new created instance in the
    #  %database.
    #
    #  Empty keyword arguments will be ignored.
    #
    #  \see _setAttrs()
    def __init__(self, **kwargs):
        self._setAttrs(True, **kwargs)


    ## \brief Convert any value into a unicode string
    #
    #  This function tries to convert any value into a unicode string.
    #  Character strings will be converted directly into unicode
    #  strings. Non-string types will be converted into a character string
    #  first.
    #
    #  \param self  The object instance itself
    #  \param value The value
    #  \return      A unicode string or \c None
    #
    #  \test tests.test_objects.TestObjects.test_AttributeConversion()
    def _toUnicode(self, value):
        # return value if is already a unicode string
        if type(value) in [types.UnicodeType, types.NoneType]:
            return value

        # try to convert into a character string
        if type(value) != types.StringType:
            try:
                _value = str(value)
            except:
                logging.warning(
                    _(u'Umwandlung von "%(value)s" (%(type)s) in eine Zeichenkette schlug fehl!'),
                    {
                        'value': value,
                        'type':  type(value),
                    })
                _value = None
        else:
            _value = value

        # convert character string into a unicode string 
        _value = unicode(_value, self.encoding)
        return _value


    ## \brief Convert any value into a decimal.Decimal value
    #
    #  \param self  The object instance itself
    #  \param value The value
    #
    #  \test tests.test_objects.TestObjects.test_AttributeConversion()
    def _toDecimal(self, value):
        # return value if is already a Decimal class
        if isinstance(value, decimal.Decimal) or type(value) == types.NoneType:
            return value

        # direct conversion from integer and long works fine
        if type(value) in [types.IntType, types.LongType]:
            _value = value
        # convert via character string
        else:
            try:
                _value = str(value)
                _value = decimal.Decimal(_value)
            except decimal.InvalidOperation:
                logging.warning(
                    _(u'Umwandlung von "%(value)s" (%(type)s) in eine Dezimalzahl schlug fehl!'),
                    {
                        'value': value,
                        'type':  type(value),
                    })
                return None

        # convert character string into a unicode string 
        _value = decimal.Decimal(_value)
        return _value


    ## \brief Update arguments (really)
    #
    #  The keyword arguments will check for completeness using the 
    #  \ref arguments set.
    #
    #  Attributes will be converted automatically. See \ref arguments
    #  for more details. Empty unicode strings will be set to \c None.
    #  Leading and trailing whitespaces will be removed.
    #
    #  \param self        The object instance itself
    #  \param ignoreEmpty Ignore attributes that are \c None or an empty string
    #  \param kwargs      Keyword arguments
    #
    #  \exception ValueError For unknown arguments
    #
    #  \test tests.test_objects.TestObjects.test_setAttrs()
    def _setAttrs(self, ignoreEmpty, **kwargs):
        current_args = set(kwargs)
        known_args   = set(self.arguments)

        # check for unknown arguments
        unknown_args = current_args.difference(known_args)
        if unknown_args:
            raise ValueError, "Unknown arguments %s found" % \
                ", ".join(unknown_args)

        # set parameters
        for attr in kwargs:
            # skip empty attributes
            value = kwargs[attr]
            if ignoreEmpty and not value:
                continue

            # convert and set attributes
            attr_type = self.arguments.get(attr)

            if attr_type == 'unicode':
                # set empty unicode strings to none
                if value:
                    value = self._toUnicode(value)
                    value = value.strip()
                if not value:
                    value = None

            elif attr_type == 'decimal':
                value = self._toDecimal(value)

            else:
                value = kwargs[attr]
            setattr(self, attr, value)

    ## \brief Returns the representation of this object as string.
    def __repr__(self):
        return self._encode(
            "<%s object at 0x%x -- %s>",
            self.__class__.__name__,
            id(self),
            self.__str__()
            )

    ## \brief Convert message and arguments to the default encoding.
    #
    #  Convert the message text as well as all unicode arguments to the 
    #  to the encoding given in \ref encoding.
    #
    #  \note Please use either \c args or \c kwargs. Not both together!
    #
    #  \param self    The object instance itself
    #  \param msg     Message text
    #  \param args    Argument list
    #  \param kwargs  Keyword arguments
    #  \return        The completed and encoded text
    #  \see           \ref encoding
    def _encode(self, msg, *args, **kwargs):
        # convert the message to the default encoding
        if type(msg) == types.UnicodeType:
            _msg = msg.encode(self.encoding, 'replace')
        else:
            _msg = msg
    
        # convert elements of the argument lists to the default encoding
        if args:
            _args = list(args[:])
            for pos in range(len(_args)):
                if type(_args[pos]) == types.UnicodeType:
                    _args[pos] = _args[pos].encode(self.encoding, 'replace')
            return _msg % tuple(_args)
    
        # convert elements of the keyword arguments to the default encoding
        elif kwargs:
            _kwargs = kwargs.copy()
            for key in _kwargs:
                if type(_kwargs[key]) == types.UnicodeType:
                    _kwargs[key] = _kwargs[key].encode(self.encoding, 'replace')
            return _msg % _kwargs
    
        # return the encode message only
        return _msg


    ## \brief Update instance arguments
    #
    #  Empty keyword arguments are processed. It means that single data will
    #  be deleted from the %database.
    #
    #  \see _setAttrs()
    def update(self, **kwargs):
        self._setAttrs(False, **kwargs)


    ## \brief Add an object to the %database and commit open changes
    #
    #  \param self   The object instance itself
    #  \param commit Make changes persistent by committing them
    #
    #  \see database.store
    def add2db(self, commit=True):
        database.store.add(self)
        if commit:
            database.store.commit()


## \brief Class to handle Exchange data
#
#  \test tests.test_objects.TestObjects.test_Exchange_objects()
class Exchange(SMMBaseObject):

    arguments  = {
        'symbol':      'unicode',
        'description': 'unicode',
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_exchange"

    ## Unique ID for each object
    id          = Int(primary=True)

    ## Unique symbol for each exchange
    symbol      = Unicode()

    ## The full name and country of the exchange
    description = Unicode()

    ## \brief Returns the content of this instance as string
    def __str__(self):
        return self._encode(
            "%s (#%s, desc: %s)",
            self.symbol,
            self.id,
            self.description
            )

    ## \brief Query Exchange %objects
    #
    #  There are three ways for querying %objects:
    #   - Querying by \c id returns a Exchange object or \c None
    #   - Querying by \c symbol returns a Exchange object or \c None
    #   - Querying without parameters to get all Exchange %objects
    #
    #  The last way returns always a storm.store.ResultSet.
    #
    #  \param id      Returns a Exchange specified by \c id.
    #  \param symbol  Returns a Exchange specified by \c symbol.
    #  \param city    Returns a Exchange specified by \c city.
    #
    #  \par Searching for city:
    #  Searching to a exchange specified by the \c city parameter
    #  searches the database currently SQLite using the \c LIKE
    #  operator.
    #
    #  To balance the diffenent behaviour of LIKE between ascii and
    #  unicode sequences. The \c city parameter is always capitalised.
    #  
    #  \par SQLite documentation of the \c LIKE operator:
    #  The default behavior of the LIKE operator is to ignore case for
    #  ASCII characters. Hence, by default 'a' LIKE 'A' is true. The
    #  case_sensitive_like pragma installs a new application-defined
    #  LIKE function that can change this behavior. When
    #  case_sensitive_like is enabled, 'a' LIKE 'A' is false but 'a'
    #  LIKE 'a' is still true.
    #
    #  \par Sorting:
    #  The result are unsorted
    #
    #  \test tests.test_objects.TestObjects.test_Exchange_get()
    @staticmethod
    def get(id=None, symbol=None, city=None):
        # query store
        if id:
            return database.store.get(Exchange, id)

        if symbol:
            if type(symbol) != types.UnicodeType:
                sym = unicode(symbol, 'latin-1')
            else:
                sym = symbol
            result_set = database.store.find(Exchange, Exchange.symbol == sym)
            if result_set.count():
                return result_set[0]
            return None

        if city:
            if type(city) != types.UnicodeType:
                _city = unicode(city, 'latin-1')
            else:
                _city = city
            result_set = database.store.find(
                Exchange,
                Exchange.description.like(
                    u'%%%s%%' % city.capitalize()
                    )
                )
            if result_set.count():
                return result_set[0]
            return None

        result_set = database.store.find(Exchange)
        return result_set

    ## \brief Returns the number of known \ref Exchange "exchanges"
    @staticmethod
    def getCount():
        res = database.store.find(Exchange)
        return res.count()


## \brief Class to handle Currency data
#
#  \test tests.test_objects.TestObjects.test_Currency_objects()
class Currency(SMMBaseObject):

    arguments  = {
        'code':        'unicode',
        'sign':        'unicode',
        'description': 'unicode',
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_currency"

    ## Unique ID for each object
    id   = Int(primary=True)

    ## ISO 4217 currency code
    code = Unicode()

    ## Short sign
    sign = Unicode()

    ## Detailed description
    description = Unicode()

    ## \brief Returns the content of this instance as string
    def __str__(self):
        return self._encode(
            "%s (#%s)",
            self.code,
            self.id,
            )

    ## \brief Query Currency %objects
    #
    #  There are two ways for querying %objects:
    #   - Querying by \c id returns a Currency object or \c None
    #   - Querying by \c code returns a Currency object or \c None
    #   - Querying by \c sign returns a Currency object or \c None
    #   - Querying without parameters to get all Currency %objects
    #
    #  The last way returns always a storm.store.ResultSet.
    #
    #  \param id    Returns a Currency %object specified by \c id
    #  \param code  Returns a Currency %object specified by \c code
    #  \param sign  Returns a Currency %object specified by \c sign
    #
    #  \par Sorting:
    #  The result are unsorted
    #
    #  \test tests.test_objects.TestObjects.test_Currency_get()
    @staticmethod
    def get(id=None, code=None, sign=None):
        if id:
            return database.store.get(Currency, id)

        expressions = []
        if code:
            if type(code) != types.UnicodeType:
                co = unicode(code, 'latin-1')
            else:
                co = code
            expressions += [ Currency.code == co ]

        if sign:
            expressions += [ Currency.sign == sign ]

        result_set = database.store.find(Currency, *expressions)

        if code or sign:
            if result_set.count():
                return result_set[0]
            return None

        return result_set

    ## \brief Returns the number of known \ref Currency "currencies"
    @staticmethod
    def getCount():
        res = database.store.find(Currency)
        return res.count()


## \brief Class to handle share data
#
#  \test tests.test_objects.TestObjects.test_Share_objects()
class Share(SMMBaseObject):

    arguments  = {
        'symbol':      'unicode',
        'isin':        'unicode',
        'wkn':         'unicode',
        'name':        'unicode',
        'currency':    None,
        'currency_id': None,
        'exchange':    None,
        'exchange_id': None,
        'kind':        'unicode',
        'watch':       None,
        'watch_id':    None,
        'source':      None,
        'notes':       'unicode',
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_share"

    ## Unique ID for each object
    id = Int(primary=True)

    ## Ticker symbol of the share
    symbol = Unicode()

    ## ISIN of the share
    isin = Unicode()

    ## WKN of the share
    wkn = Unicode()

    ## Name of the company
    name = Unicode()

    ## ID of the referred Currency %object
    currency_id = Int()

    ## \brief Preferred currency
    #
    #  %Reference to the Currency %object
    currency = Reference(currency_id, Currency.id)

    ## ID of the referred Exchange %object
    exchange_id = Int()

    ## \brief Preferred exchange
    #  
    #  %Reference to the Exchange %object
    exchange = Reference(exchange_id, Exchange.id)

    ## Type of share
    #
    #  Adapt functions below after adding a new kind:
    #  - gui.helper.formatShareKind()
    #  - gui.controls.ShareKindChoice.__init__()
    kind = Enum(map={'share': 1, 'index': 2, 'fond': 3, 'etf': 4, \
                     'warrant': 5, 'certificate': 6, 'bond': 7})

    ## ID of the reffered watch list entry
    watch_id = Int()

    ## %Reference to the Watch list %object
    watch = Reference(watch_id, 'Watch.id')

    ## Primary data source
    source = Int()

    ## Additional notes to this share
    notes = Unicode()

    ## Create a watch list entry (Watch %object) releated to the current share.
    #
    #  \param self   The object instance itself
    #  \param kwargs Same like Watch.__init__(). The \c share parameter is
    #                set automatically.
    #
    #  \return \c False if a watch list entry is already referred,
    #          otherwise \c True.
    def createWatch(self, **kwargs):
        if self.watch:
            return False
        kwargs['share'] = self
        self.watch = Watch(**kwargs)
        self.watch.add2db()
        return True

    ## \brief Returns the content of this instance as string
    def __str__(self):
        if self.isin:
            isin = ' isin: %s' % self.isin
        elif self.wkn:
            isin = ' wkn: %s' % self.wkn
        elif self.symbol:
            isin = ' symbol: %s' % self.symbol
        else:
            isin = ''
        return self._encode('%s (#%s, %s)', self.name, self.id, isin)

    ## \brief Returns \c True if a related watch list entry exists.
    #
    #  Use this function only if you need an explicit \c True or \c
    #  False.
    #
    #  You can check the \c watch attribute directly as an alternative.
    #
    #  Or you can convert the \c watch attribute to a boolean value:
    #  \code
    #  bool(share.watch)
    #  \endcode
    def is_watched(self):
        return bool(self.watch)

    ## \brief Query Share %objects
    #
    #  There are three ways for querying %objects:
    #   - Querying by \c id, \c isin, \c symbol or \c wkn returns a Share
    #     object or \c None
    #   - Querying by \c kind returns all Share %objects with the given
    #     kind
    #   - Querying without parameters to get all Share %objects
    #
    #  The last way returns always a storm.store.ResultSet.
    #
    #  \param id      Returns a Share specified by \c id.
    #  \param isin    Returns a Share specified by \c isin.
    #  \param symbol  Returns a Share specified by \c symbol.
    #  \param wkn     Returns a Share specified by \c wkn.
    #  \param index   Returns a Indices
    #
    #  \par Sorting:
    #  The result are unsorted
    #
    #  \test tests.test_objects.TestObjects.test_Share_get()
    @staticmethod
    def get(id=None, isin=None, symbol=None, wkn=None, index=None):
        # query store
        if id:
            return database.store.get(Share, id)

        expressions = []
        if isin:
            if type(isin) != types.UnicodeType:
                _isin = unicode(isin, 'latin-1')
            else:
                _isin = isin
            expressions += [ Share.isin == _isin ]
        if symbol:
            if type(symbol) != types.UnicodeType:
                _sym = unicode(symbol, 'latin-1')
            else:
                _sym = symbol
            expressions += [ Share.symbol == _sym ]
        if wkn:
            if type(wkn) != types.UnicodeType:
                _wkn = unicode(wkn, 'latin-1')
            else:
                _wkn = wkn
            expressions += [ Share.wkn == _wkn ]

        if index:
            expressions += [ Share.kind == 'index' ]
        else:
            expressions += [ Share.kind != 'index' ]

        result_set = database.store.find(Share, *expressions)

        if isin or symbol or wkn:
            if result_set.count():
                return result_set[0]
            else:
                return None

        return result_set

    ## \brief Returns the number of shares, fonds and others
    #
    #  \param fonds Returns the number of known fonds
    #  \param index Returns the number of known indexes
    #  \param other Returns the number of not catagorised instruments
    #  \param share Returns the number of known shares
    #  \param total Returns the number of instruments wihtout indexes
    @staticmethod
    def getCount(fonds=False, index=False, other=False, share=False, total=False):
        expressions = []
        if fonds:
            expressions += [ Share.kind == 'fond' ]
        if index:
            expressions += [ Share.kind == 'index' ]
        if other:
            count = Share.getCount(total=True) - \
                    Share.getCount(share=True) - \
                    Share.getCount(fonds=True)
            return count
        if share:
            expressions += [ Share.kind == 'share' ]
        if total:
            count = Share.getCount() - Share.getCount(index=True)
            return count

        res = database.store.find(Share, *expressions)
        return res.count()

    ## \brief Returns the preferred currency of a share.
    #
    #  The preferred currency is the currency of a current open Position or
    #  the currency stored at the share object.
    #
    #  \c None is returned if the currency can't determinate.
    #
    #  \return A smm.objects.Currency %object or \c None
    def getPrefCurrency(self):
        pos = Position.get(share=self, open=True)
        if pos:
            return(pos.currency)
    
        if self.currency:
            return(self.currency)
    
        return None


## \brief Class to handle quotes 
#
#  \note This class handles daily quotes only. Another class will handle
#        intraday quotes.
#
#  \note There is not field to track share splits as well as dividend
#        currently.
#
#  \test tests.test_objects.TestObjects.test_DailyQuote_objects()
class DailyQuote(SMMBaseObject):

    arguments  = {
        'share_id':    None,
        'share':       None,
        'exchange_id': None,
        'exchange':    None,
        'currency_id': None,
        'currency':    None,
        'open':        'decimal',
        'close':       'decimal',
        'high':        'decimal',
        'low':         'decimal',
        'date':        None,
        'volume':      'decimal',
        'source':      None,
        'event':       'unicode',
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_dailyquote"

    ## Unique ID for each object
    id = Int(primary=True)

    ## Property to store the ID of the referred Share %object
    share_id = Int()

    ## %Reference to the Share %object
    share = Reference(share_id, Share.id)

    ## Property to store the ID of the referred Exchange %object
    exchange_id = Int()

    ## %Reference to the Exchange %object
    exchange = Reference(exchange_id, Exchange.id)

    ## Property to store the ID of the referred Currency %object
    currency_id = Int()

    ## %Reference to the Currency %object
    currency = Reference(currency_id, Currency.id)

    ## Opening value of the share
    open = Decimal()

    ## Closing value of the share
    close = Decimal()

    ## Highest value of the day
    high = Decimal()

    ## Lowest value of the day
    low = Decimal()

    ## Day of quotation
    date = Date()

    ## Trade volume in shares
    volume = Int()

    ## Data source
    source = Int()

    ## Event releated to this quote
    #
    #  It's more like a note ...
    #
    #  Currently supported events:
    #   - split \code split:1:2 \endcode
    #   - bid - bid price (German: Geldkurs)
    #   - ask - ask price (German: Briefkurs)
    #
    #  Adapt functions below to add a new event:
    #  - gui.helper.formatEvent()
    event  = Unicode()


    ## \brief Returns the content of this instance as string
    def __str__(self):
        return self._encode(
           'DailyQuote:\n' \
           '  id:       %(id)s\n'  \
           '  date:     %(date)s\n' \
           '  share:    %(share)s\n' \
           '  exchange: %(exchange)s\n' \
           '  currency: %(currency)s\n' \
           '  open:     %(open)s\n' \
           '  close:    %(close)s\n' \
           '  low:      %(low)s\n' \
           '  high:     %(high)s\n' \
           '  volume:   %(volume)s\n',
           id=       self.id, 
           share=    self.share,
           exchange= self.exchange,
           currency= self.currency,
           low=      self.low,
           high=     self.high,
           date=     self.date,
           open=     self.open,
           close=    self.close,
           volume=   self.volume
           )

    ## \brief Query DailyQuote %objects
    #
    #  There are three ways for querying %objects:
    #   - Querying by \c id returns a DailyQuote object or \c None
    #   - Query special entries described using \c share and/or \c date
    #   - Querying without parameters to get all DailyQuote %objects
    #
    #  Querying with \c share and \c date returns an DailyQuote object or
    #  \c None.
    #
    #  The last two ways returns always a storm.store.ResultSet.
    #
    #  \param id    Returns a quote specified by the id.
    #  \param share Returns quotes specified by share.
    #  \param date  Returns a quote specified by date.
    #
    #  \par Sorting:
    #  The result is ordered ascending by date
    #
    #  \test tests.test_objects.TestObjects.test_DailyQuote_get()
    @staticmethod
    def get(date=None, id=None, share=None):
        if id:
            return database.store.get(DailyQuote, id)

        expressions = []
        if date:
            expressions += [ DailyQuote.date == date ]
        if share:
            expressions += [ DailyQuote.share == share ]

        result_set = database.store.find(DailyQuote, *expressions)
        result_set = result_set.order_by(Asc(DailyQuote.date))

        if share and date:
            if result_set.count():
                return result_set[0]
            return None
        return result_set

    ## \brief Returns the number of \ref DailyQuote "daily quotes"
    @staticmethod
    def getCount():
        res = database.store.find(DailyQuote)
        return res.count()

    ## \brief Returns the latest DailyQuote for a share
    #
    #  \param share    Querying for this Share %object
    #  \param currency Specifiy the currency for the DailyQuote
    #
    #  \test tests.test_objects.TestObjects.test_DailyQuote_getLast()
    @staticmethod
    def getLast(share, currency=None):
        expressions = [DailyQuote.share == share]
        if currency:
            expressions += [DailyQuote.currency == currency]
        result_set = database.store.find(DailyQuote, *expressions)
        return result_set.order_by(Desc(DailyQuote.date)).first()


## \brief Class to handle transations
#
#  There are different kinds of transactions with different set of
#  mandatory and non-mandatory attributes.
#
#  The attributes \ref count, \ref date, \ref kind and \ref value are
#  mandatory generally. The attributes \ref exchange and \ref fee are
#  always non-mandatory. The \ref date attribute is a small exception
#  because it will be set to the current day if it's not given.
#
#  The \c split transaction is an other exception caused by the nature
#  of this transaction type. The attribute \ref value is non-mandatory.
#
#  \test tests.test_objects.TestObjects.test_Transaction_objects()
class Transaction(SMMBaseObject):

    arguments  = {
        'count':       'decimal',
        'value':       'decimal',
        'fee':         'decimal',
        'date':        None,
        'exchange':    None,
        'exchange_id': None,
        'kind':        'unicode',
        'position':    None,
        'position_id': None,
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_transaction"

    ## Unique ID for each object
    id = Int(primary=True)

    ## Number of shares to buy or sell
    count = Decimal()

    ## Value of the share
    value = Decimal()

    ## Overall transation fee
    fee = Decimal()

    ## Date of the transaction
    date = Date()

    ## Property to store the ID of the referred Exchange %object
    exchange_id = Int()

    ## %Reference to the Exchange %object
    exchange = Reference(exchange_id, Exchange.id)

    ## Type of transaction
    #
    #  Adapt functions below after adding a new kind:
    #  - Transaction.getVolume()
    #  - gui.helper.formatTransactionKind()
    #  - gui.controls.TransactionKindChoice.__init__()
    kind = Enum(map={'buy': 1, 'sell': 2, 'dividend': 3, 'split': 4})

    ## Property to store the ID of the referred Position %object
    position_id = Int()

    ## %Reference to the Position %object
    position = Reference(position_id, 'Position.id')

    ## Create a new Transaction instance
    #
    #  \param self      The object instance itself
    #  \param count     Number of shares to trade (\c decimal.Decimal)
    #  \param date      Transaction.date (\c datetime.date)
    #  \param exchange  Exchange to trade
    #  \param fee       Transaction.fee (\c decimal.Decimal)
    #  \param kind      Transaction.kind
    #  \param value     Value of a single share (\c decimal.Decimal)
    #  \param position  Position %object
    #
    #  \exception ValueError Raised for missing arguments
    def __init__(self, count, kind, position, date=datetime.date.today(), \
                 exchange=None, fee=None, value=None):

        if kind != 'split' and not value:
            raise ValueError, "Missing argument: value"

        SMMBaseObject.__init__(
            self,
            count=count,
            date=date,
            exchange=exchange,
            fee=fee,
            kind=kind,
            value=value,
            position=position
            )

    ## \brief Returns the content of this instance as string
    def __str__(self):
        return self._encode(
          '#%(id)s (date: %(date)s, count: %(count)s, value: %(value)s, kind: %(kind)s, position: %(position)s)',
           id=       self.id, 
           date=     self.date, 
           count=    self.count,
           value=    self.value,
           kind=     self.kind,
           position= self.position,
           )

    ## \brief Query Transaction %objects
    #
    #  \param id    Returns a single transaction specified by id.
    #
    #  \par Sorting:
    #  The result is ordered ascending by date
    #
    #  \test tests.test_objects.TestObjects.test_Transaction_get()
    @staticmethod
    def get(id=None):
        if id:
            return database.store.get(Transaction, id)

        result_set = database.store.find(Transaction)
        result_set = result_set.order_by(Asc(Transaction.date))
        return result_set

    ## \brief Returns the total number of Transaction
    @staticmethod
    def getCount():
        res = database.store.find(Transaction)
        return res.count()

    ## \brief Returns all Transaction between \c start and \c end date
    #
    #  \param start Start date
    #  \param end   End date
    #  \param pos   Return transactions associated with this Position
    @staticmethod
    def getInPeriod(start, end, pos=None):
        expressions = [
           Transaction.date >= start,
           Transaction.date <= end,
           ]
        if pos:
            expressions += [ Transaction.position_id == pos.id ]
        result_set = database.store.find(Transaction, *expressions)
        result_set = result_set.order_by(Asc(Transaction.date))
        return result_set

    ## \brief Calculate the transaction volume
    #
    #  Aquisitions increased the volume. Sales decreases it. Dividends are
    #  added to the volume. Splits doesn't have any effect.
    #
    #  Fees are processed for buy and sell transactions only.
    #
    #  \param self    The object instance itself
    #  \param withFee Include the transaction fees in the volume
    #
    #  \return        Volume of the transactions as \c decimal.Decimal value. 
    #                 Split transactions returns \c decimal.Decimal('0').
    #
    #  \test tests.test_objects.TestObjects.test_Transaction_getVolume()
    def getVolume(self, withFee=True):
        # ignore splits
        if self.kind == 'split':
            return decimal.Decimal(0)

        volume = self.count * self.value
        if withFee and self.fee:
            if self.kind == 'sell':
                volume -= self.fee
            elif self.kind == 'buy':
                volume += self.fee
            else: 
                logging.warning(
                    _(u'Unerwartete Gebhr in Transaktion #%d'),
                    self.id
                    )

        return volume

    ## \brief Returns the holding period of shares.
    #
    #  The function calculates the difference between a transaction date
    #  and given date.
    #
    #  \param self  The object instance itself
    #  \param date  Date of the end of the holding period
    #  \return      The duration in days
    def getHoldingPeriod(self, date=datetime.date.today()):
        delta = date - self.date
        return delta.days
    

## \brief Class to handle watch list entries
#
#  \test tests.test_objects.TestObjects.test_Watch_objects()
class Watch(SMMBaseObject):

    arguments  = {
        'created':  None,
        'expire':   None,
        'notes':    'unicode',
        'share':    None,
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_watch"

    ## Unique ID for each object
    id   = Int(primary=True)

    ## Detailed notes
    notes = Unicode()

    ## \brief %Reference to the Share %object
    #
    #  This uses a backwards references to prevent trouble with cross
    #  references
    share = Reference(id, Share.watch_id, on_remote=True)

    ## Date when the watch object is created
    created = Date()

    ## Date when the watch object will be removed automatically
    expire = Date()

    ## Create a new Watch instance
    #
    #  \param self     The object instance itself
    #  \param share    Share to watch (smm.objects.Share %object)
    #  \param created  Creation date (datetime.date %object, default: today)
    #  \param expire   Expiration date (datetime.date %object, default: no expiration)
    #  \param notes    Comment to the entry (default: \c None)
    def __init__(self, share, created=datetime.date.today(), expire=None,
                         notes=None):

        SMMBaseObject.__init__(
            self,
            share=share,
            created=created,
            expire=expire,
            notes=notes
            )


    ## \brief Returns the content of this instance as string
    def __str__(self):
        return self._encode(
           '#%(id)s (created: %(created)s, expire: %(expire)s, share: %(share)s)',
           id      = self.id,
           created = self.created,
           expire  = self.expire,
           share   = self.share,
            )

    ## \brief Query Watch %objects
    #
    #  There are three ways for querying %objects:
    #   - Querying by \c id returns a Watch object or \c None
    #   - Query special entries described using \c share and/or \c expired
    #   - Querying without parameters to get all Watch %objects
    #
    #  The last two ways returns always a storm.store.ResultSet.
    #
    #  \param id      Returns a watch list entry specified by id.
    #  \param expired Boolean flag to returns all expired watch list entries.
    #  \param share   Returns a watch list entry specified by share.
    #
    #  \par Sorting:
    #  The result are unsorted
    #
    #  \test tests.test_objects.TestObjects.test_Watch_get()
    @staticmethod
    def get(id=None, expired=False, share=None):
        # query store
        if id:
            return database.store.get(Watch, id)

        expressions = []
        if share:
            expressions += [ Watch.share == share ]
        if expired:
            expressions += [ Watch.expire < datetime.date.today() ]

        result_set = database.store.find(Watch, *expressions)
        return result_set

    ## \brief Delete watch list entry
    #
    #  The reference between share and watch list entry will be cleaned
    #  and Watch object will be removed from the %database.
    def remove(self):
        self.share.watch = None
        database.store.remove(self)
        database.store.commit()
        

## \brief Class to handle watch list entries
#
#  \test tests.test_objects.TestObjects.test_Position_objects()
class Position(SMMBaseObject):

    arguments  = {
        'currency_id': None,
        'currency':    None,
        'notes':       'unicode',
        'open':        None,
        'share_id':    None,
        'share':       None,
        }

    ## Use given table to store all of this class inside
    __storm_table__ = "t_position"

    ## Unique ID for each object
    id   = Int(primary=True)

    ## ID of the referred Currency %object
    currency_id = Int()

    ## %Reference to the Currency %object
    currency = Reference(currency_id, Currency.id)

    ## Additional notes to this position
    notes = Unicode()

    ## Flag to mark open positions 
    open = Bool(default=True)

    ## Property to store the ID of the referred Share %object
    share_id = Int()

    ## %Reference to the Share %object
    share = Reference(share_id, Share.id)

    ## Property to store the ID of the referred Stop %object
    stop_id = Int()

    ## %Reference to all Transaction associated to this position
    transactions = ReferenceSet(
        id,
        Transaction.position_id
        )

    ## \brief Returns the content of this instance as string
    def __str__(self):
        if self.open:
            pos_open = 'open'
        else:
            pos_open = 'closed'
        return self._encode(
           'Position for %(name)s (#%(id)s, currency: %(currency)s, %(open)s, share: %(share)s)',
           name     = self.share.name,
           id       = self.id,
           currency = self.currency,
           open     = pos_open,
           share    = self.share,
            )

    ## \brief Query Position %objects
    #
    #  There are three ways for querying %objects:
    #   - Querying by \c id returns a Position object or \c None
    #   - Querying entries using \c share
    #   - Querying without parameters to get all Position %objects
    #
    #  Querying with \c share and \c open returns an Position object or
    #  \c None.
    #
    #  The last two ways returns always a storm.store.ResultSet.
    #
    #  \param id       Returns a Position specified by \c id.
    #  \param open     Returns all open / closed \ref Position "Positions"
    #  \param share    Returns all Positions matched with \c share.
    #  \param currency Returns all Positions matched with \c currency.
    #
    #  \par Sorting:
    #  The result are unsorted
    #
    #  \test tests.test_objects.TestObjects.test_Position_get()
    @staticmethod
    def get(id=None, open=None, share=None, currency=None):
        # query store
        if id:
            return database.store.get(Position, id)

        expressions = []
        if open is not None:
            expressions += [ Position.open == open ]
        if share:
            expressions += [ Position.share == share ]
        if currency:
            expressions += [ Position.currency == currency ]

        result_set = database.store.find(Position, *expressions)
        if share and open:
            if result_set.count():
                return result_set[0]
            return None
        return result_set

    ## \brief Returns the number of known \ref Position "Positions"
    #
    #  \param open   Returns the number of all open Position %objects
    #  \param closed Returns the number of all closed Position %objects
    @staticmethod
    def getCount(closed=False, open=False):
        expressions = []
        if closed:
            expressions += [ Position.open == False ]
        if open:
            expressions += [ Position.open == True ]

        res = database.store.find(Position, *expressions)
        return res.count()

    ## \brief Returns all Position with Transaction between \c start and
    #         \c end date
    #
    #  \param start Start date
    #  \param end   End date
    #  \param open  Look for open Position only
    #
    #  \par Sorting:
    #  The result are unsorted
    @staticmethod
    def getInPeriod(start, end, open=False):
        expressions = [
            Position.id == Transaction.position_id,
            Transaction.date >= start,
            Transaction.date <= end,
            ]
        if open:
            expressions += [ Position.open == True ]

        result_set = database.store.find(Position, *expressions)
        result_set.group_by(Position)
        return result_set

    ## \brief Returns the numbers of shares in this position
    #  \test tests.test_objects.TestObjects.test_Position_getPerf()
    def getShareCount(self):
        count = decimal.Decimal(0)
        for transaction in self.transactions:
            # dividend transactions didn't change the number of shares
            if transaction.kind == 'dividend':
                continue
            count += transaction.count
        return count

    ## \brief Calculate the volume of a Position
    #
    #  \return Volume in \c decimal.Decimal
    #
    #  \see getTotalVolume()
    def getVolume(self):
        volume = decimal.Decimal(0)
        quote, unused = self.getLastQuote()
        if quote:
            volume = self.getShareCount() * quote
        return volume

    ## Calculate the volume of all open positions
    #
    #  \param currency Specifies the currency to summarise all open positions
    #
    #  \return Volume (\c decimal.Decimal)
    #
    #  \see getVolume()
    @staticmethod
    def getTotalVolume(currency):
        volume = decimal.Decimal(0)
        position_list = Position.get(open=True, currency=currency)
        for position in position_list:
            volume += position.getVolume()
        return volume

    ## \brief Calculates the performance of a Position
    #
    #  \param self     The object instance itself
    #  \param withFee  Calculate performance with or without
    #
    #  \returns The absolute and relative performance in a tuple.
    #  \returns Two times \c decimal.Decimal("0") for no suitable
    #           transactions.
    #
    #  \test tests.test_objects.TestObjects.test_Position_getPerf()
    def getPerf(self, withFee=False):
        # initialise variables
        #=====================
    
        # costs is negative and income is positive
        # fees parts of costs
        income = costs = decimal.Decimal(0)
        perf_abs = perf_rel = decimal.Decimal(0)
    
        # get transaction list and last quote
        #====================================
        if not self.transactions.count():
            logging.debug(
                _(u'Es wurde keine Transaktionen fr "%s" gefunden.'),
                self.share.name
                )
            return (None, None)
    
        quote, unused = self.getLastQuote()
        if not quote:
            logging.debug(
                _(u'Es wurde keine aktueller Kurs fr "%s" gefunden.'),
                self.share.name
                )
            return (None, None)
    
        # process transactions
        #=====================
        for transaction in self.transactions:
            if transaction.kind == 'buy':
                costs -= transaction.count * transaction.value
                if withFee and transaction.fee:
                    costs -= transaction.fee
            elif transaction.kind == 'dividend':
                income += transaction.count * transaction.value
            elif transaction.kind == 'split':
                pass
            elif transaction.kind == 'sell':
                income += -1 * transaction.count * transaction.value
                if withFee and transaction.fee:
                    costs -= transaction.fee
            else:
                logging.warning(
                    _(u'Unbekannter Transaktionstyp "%(kind)s" im Wertpapier "%(name)s" gefunden.'),
                    {
                        'kind': transaction.kind,
                        'name': self.share.name,
                    })
    
        # compute the performance
        value_cur = quote * self.getShareCount()
    
        # remember: costs is negative!
        perf_abs = value_cur + income + costs
        perf_rel = 100 * perf_abs / (-1 * costs)
    
        return (perf_abs, perf_rel)

    ## \brief Calculates the average buy price
    #
    #  The functions query all transactions of the current share position
    #  and calculates the <b>weighted arithmetic mean</b> of all "buy"
    #  transactions.
    #
    #  \param self  The object instance itself
    #
    #  \return Average buy as \c decimal.Decimal value
    def getAvgBuy(self):
        if not self.transactions.count():
            return decimal.Decimal('0')
    
        count = volume = decimal.Decimal('0')
        for trans in self.transactions:
            if trans.kind != 'buy':
                continue
    
            count += trans.count
            volume += trans.count * trans.value

        return volume / count

    ## \brief Returns the latest quotation of a share
    #
    #  The function returns a tuple of two values:
    #  \li the quotation (type \c decimal.Decimal)
    #  \li the date (type \c datetime.Date)
    #
    #  All return values are \c None in case of error.
    #
    #  The function returns the latest daily %quote if found.
    #  Otherwise is returns the data based on the last transaction.
    #  Or three times none, if nothing has been found.
    #
    #  Dividend transactions will be ignored.
    #
    #  \return A tuple of two values
    #
    #  \test tests.test_objects.TestObjects.test_Position_getLastQuote()
    def getLastQuote(self):
        quotation  = None
        date       = None
        quote_date = None
        trans_date = None
    
        daily_quote = DailyQuote.getLast(self.share, self.currency)
        if daily_quote:
            quote_date = daily_quote.date
    
        # search all transaction that aren't dividend payments
        result_set = self.transactions.find(
            Transaction.kind != 'dividend'
            )
        result_set.order_by(Asc(Transaction.date))

        if result_set.count():
            trans = result_set.last()
            trans_date = trans.date
    
        if quote_date and not trans_date:
            quotation = daily_quote.close
            date      = daily_quote.date
        elif not quote_date and trans_date:
            quotation = trans.value
            date      = trans.date
        elif quote_date < trans_date:
            quotation = trans.value
            date      = trans.date
        elif quote_date > trans_date:
            quotation = daily_quote.close
            date      = daily_quote.date
        else:
            quotation = daily_quote.close
            date      = daily_quote.date
    
        return (quotation, date)

    ## Buy shares for the current Position
    #
    #  \param self      The object instance itself
    #  \param count     Number of shares to trade (\c decimal.Decimal)
    #  \param value     Value of a single share (\c decimal.Decimal)
    #  \param fee       Transaction.fee (\c decimal.Decimal)
    #  \param exchange  Exchange to trade
    #  \param date      Transaction.date (\c datetime.date)
    #
    #  \returns The created transaction object
    def buy(self, count, value, fee=None, exchange=None, \
            date=datetime.date.today()):
        tr = Transaction(
            count=count,
            value=value,
            fee=fee,
            date=date,
            exchange=exchange,
            kind='buy',
            position=self
            )
        tr.add2db()
        return tr

    ## Sell parts of the current Position
    #
    #  \param self      The object instance itself
    #  \param count     Number of shares to trade (\c decimal.Decimal)
    #  \param value     Value of a single share (\c decimal.Decimal)
    #  \param fee       Transaction.fee (\c decimal.Decimal)
    #  \param exchange  Exchange to trade
    #  \param date      Transaction.date (\c datetime.date)
    #
    #  \returns The created transaction object
    def sell(self, count, value, fee=None, exchange=None, \
            date=datetime.date.today()):
        tr = Transaction(
            count=count,
            value=value,
            fee=fee,
            date=date,
            exchange=exchange,
            kind='sell',
            position=self
            )
        tr.add2db()
        return tr

    ## Add a dividend Transaction
    #
    #  \param self      The object instance itself
    #  \param count     Number of shares (\c decimal.Decimal)
    #  \param value     Value per single share (\c decimal.Decimal)
    #  \param date      Transaction.date (\c datetime.date)
    #
    #  \returns The created transaction object
    def dividend(self, count, value, date=datetime.date.today()):
        tr = Transaction(
            count=count,
            value=value,
            date=date,
            kind='dividend',
            position=self
            )
        tr.add2db()
        return tr

    ## Do a share split
    #
    #  Splitting means that the position will decreased by the number of
    #  \c old shares and incresed by the number of \c new shares. The
    #  split will be realised within two transactions.
    #
    #  \param self  The object instance itself
    #  \param old   Ratio old to new shares (\c decimal.Decimal)
    #  \param new   Ratio old to new shares (\c decimal.Decimal)
    #  \param date  Split date (\c datetime.date)
    #  \param count Number of shares to split (\c decimal.Decimal)
    #
    #  \return Both split transactions (old and new)
    def split(self, old, new, date=datetime.date.today(), count=None):
        if not count:
            count = self.getShareCount()
        count_old = -1 * old * count
        tr_old = Transaction(
            count=count_old,
            date=date,
            kind='split',
            position=self
            )
        count_new = new * count
        tr_new = Transaction(
            count=count_new,
            date=date,
            kind='split',
            position=self
            )

        # update DailyQuote.event
        dq = DailyQuote.get(share=self.share, date=date)
        if not dq:
            dq = DailyQuote(
                date=date,
                share=self.share,
                )
            dq.add2db()
        dq.event = u'split:%s:%s' % (old, new)

        database.store.add(tr_old)
        database.store.add(tr_new)
        database.store.commit()
        return tr_old, tr_new

