# -*- coding: Latin-1 -*-
#
# SMM Share and Money Management
# Copyright (c) 2007-2010 Carsten Grohmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

## \file smm/quote/base.py
#
# Basic functionality to handle finance data provided by public
# resources.
#
# \namespace smm::quote::base
# Basic functionality to handle finance data provided by public
# resources.
#
# \author Carsten Grohmann <mail (at) carstengrohmann (dot) de>
#
# \par License:
# GPL version 2.0

import datetime
import decimal
import locale
import logging
import re
import robotparser
import socket
import time
import types
import urllib2
import urlparse

import BeautifulSoup

## Base class of all quotation classes.
#
#  This class provides often used functions e.g. to access remote resources
#  or to convert date and time from strings into Python objects.
class FetchPluginBase(object):

    ## List of supported properties
    #
    #  Every class should provide a minimum set of properties
    #
    #  \li \c date     Time tuple with the date and time (if available) of
    #                  the %quote
    #  \li \c errormsg As unicode string or \c None
    #  \li \c quote.price As ascii string
    #  \li \c source   Source of the data as unicode string
    #  \li \c success  Flag as boolean value
    properties = []

    ## Property names
    _specific_properties   = []

    ## Name of the general properties
    _general_properties = ['errormsg', 'success']

    ## Default encoding of all returned strings
    encoding = 'ascii'

    ## Version number of the interface to access and convert the data
    interface_version = 2

    ## Name of the instance logger
    logger_name = 'smm.quote.base'

    ## Name of the data source
    name = 'Override in derivated class!'

    ## Unique provider ID
    provider_id = None

    ## Decimal separator for converting strings to decimal values
    #  \see attr2decimal()
    #  \see tsep
    dsep = u','

    ## Thousands separator for converting strings to decimal values
    #  \see attr2decimal()
    #  \see dsep
    tsep = None

    ## \var logger
    #  Instance logger

    ## Number of retrys for temporary connection %errors
    #
    #  \see wait
    max_retries = 3

    ## Define a own user agent string
    #
    #  Set to \c None for disable changing user agent
    user_agent = 'Mozilla/5.0 (X11; U; Linux i686; de; rv:1.9.1.6) Gecko/20091201 Firefox/3.5.6'

    ## Pre compiled RE to remove useless data from the results
    REC_REMOVE = re.compile(
        r'(&nbsp;)'        # protected whitespaces
        r'|'
        r'(&plusmn;)'      # plus minus
        r'|'
        r'(\.)'            # thousands separator "."
        r'|'
        r'(\%?\s*$)'       # tailing %
        r'|'
        r'(\s*\+?)'        # + sign'
        )

    ## RE pattern to search / match exchange symbols
    REP_EXCHANGE = r'(?P<ex_symbol>[A-Z]{1,3})'

    ## RE pattern to search / match ISIN
    #
    #  \note Compile with \c re.VERBOSE
    REP_ISIN = r'''
        (?P<isin>
        [a-zA-Z]{2}    # country code 
        [a-zA-Z0-9]{9} # National Securities Identifying Number (NSIN)
        [0-9]          # check digit 
        )
        '''

    ## RE pattern to search / match ticker symbols
    REP_SYMBOL = r'(?P<symbol>[A-Z0-9]{1,4})'

    ## RE pattern to search / match German WKN
    REP_WKN = r'(?P<wkn>[a-zA-Z0-9]{6})'

    ## \var _robots_cache
    #  Cache for already read and parsed \c robots.txt files.

    ## Time to wait beween two retries in seconds
    #
    #  \see max_retries
    wait = 60

    ## Constructor
    def __init__(self):
        self.properties = self._specific_properties[:]
        self.properties.extend(self._general_properties)
        self.properties = list(set(self.properties))
        self._robots_cache = {}
        self.logger = logging.getLogger(self.logger_name)

    ## Fetch quotes of all shares from the data provider
    #
    #  \param self       The object instance itself
    #  \param share_list List of share symbols
    #
    #  \return A dictionary structure with the share name as key to a
    #          dictionary with all queried properties. The returned
    #          properties are strings.
    #
    #  \see properties
    def getCurrent(self, share_list):
        raise NotImplementedError, \
	    "Implement this function in all derivated classes!"

    ## Return a share name based on the share details 
    #
    #  \param self  The object instance itself
    #  \param share Dictionary with the share information
    def _getName(self, share):
        if 'share.isin' in share:
            detail = _(u'ISIN %s' % share['share.isin'])
        elif 'share.wkn' in share:
            detail = _(u'WKN %s' % share['share.wkn'])
        elif 'share.symbol' in share:
            detail = _(u'Symbol %s' % share['share.symbol'])
        else:
            detail = u''

        if 'share.name' in share:
            if detail:
                res = _(u'%s (%s)' % (share['share.name'], detail))
            else:
                res = _(u'%s' % share['share.name'])
        else:
            if detail:
                res = u'(%s)' % detail
            else:
                res = _(u'Unbekannt: %s' % str(share))

        return res

    ## Cleanup the HTML code and create a BeautifulSoup.BeautifulSoup instance
    #
    #  \param self The object instance itself
    #  \param html HTML code as String
    #
    #  \return A BeautifulSoup.BeautifulSoup instance
    def getSoup(self, html):
        # remove useless characters
        _html = html.replace('&nbsp;', ' ')
        _html = [line.strip() for line in _html.split('\n')]
        _html = ''.join(_html)

        soup = BeautifulSoup.BeautifulSoup(_html, markupMassage=False)

        return soup

    ## Extract the content (text) of the next element.
    #
    #  The given parameter has to be a instance of
    #  \c BeautifulSoup.NavigableString and the next element has to be a
    #  instance of \c BeautifulSoup.Tag.
    #
    #  \param self       The object instance itself
    #  \param NavString  BeautifulSoup.NavigableString
    #  \param trim       Cut off unused characters / sequences (see
    #                    \ref REC_REMOVE)
    #
    #  \return Text of the next element or \c None in error case
    def getNextTagText(self, NavString, trim=False):
        if not isinstance(NavString, BeautifulSoup.NavigableString):
            self.logger.debug(
                _(u'Element ist keine Instanz von NavigableString sondern: %s'),
                NavString,
                )
            return None
        next = NavString.next
        if not isinstance(next, BeautifulSoup.Tag):
            self.logger.debug(
                _(u'Das nchste Element ist keine Instanz von BeautifulSoup.Tag sondern: %s'),
                type(next),
                )
            return None

        text = next.text
        # like _trim() but without removing "."
        if trim:
            text = self.REC_REMOVE.sub('', text)

        return text

    ## Return an empty initialised share dict
    #
    #  \param self   The object instance itself
    def initShareStruct(self):
        struct = {}
        struct['quote.price'] = None
        struct['errormsg']    = None
        struct['success']     = False
        struct['source.name'] = self.name
        struct['source.id']   = self.provider_id
        return struct

    ## Query a remote HTML page
    #
    #  Initialise the share specific structure using \c key
    #  (see initShareStruct()) and query a HTML page from the given \c url.
    #
    #  Querying is successful if \c pattern is found in the text.
    #
    #  In case of error contains \c struct[key]['errormsg'] a detailed error
    #  message.
    #
    #  \param self    The object instance itself
    #  \param struct  Share dict
    #  \param key     Key to store share related values in \c struct
    #  \param url     URL to open
    #  \param pattern Sinple text pattern
    #
    #  \return HTML as string or \c None
    #
    #  \see initShareStruct()
    def getHTML(self, struct, key, url, pattern):
        # initialise loop internal variables (per ISIN)
        struct[key] = self.initShareStruct()
        stock = struct[key]

        # query data
        html, errormsg = self.getRemotePage(url)
        if not html:
            stock['success']  = False
            stock['errormsg'] = errormsg
            return None

        if pattern and html.find(pattern) == -1:
            stock['success'] = False
            if not stock['errormsg']:
                stock['errormsg'] = _(u'Es wurden keine Daten zurckgeliefert.')
            return None

        return html

    ## Return the content of a remote page
    #
    #  The function fetchs a remote page and returns the content as a
    #  string. The content is \c None in case of errors and a
    #  explanatory messages is returned.
    #
    #  \param self The object instance itself
    #  \param url  URL to open
    #
    #  \return A tuple of page content and error message
    #  \see \ref getConnection()
    def getRemotePage(self, url):
        conn, errormsg = self.getConnection(url)
        if not conn:
            return (None, errormsg)
        content = conn.read()
        conn.close()
        return (content, None)

    ## Return a connection object to a remote page
    #
    #  The function opens a connection to a remote page and returns the
    #  the connection object or \c None. A explanatory messages is
    #  returned in case of %errors.
    #
    #  The request will be repeated on temporary %errors (socket.EAI_AGAIN).
    #  \ref wait and \ref max_retries controls this behaviour.
    #
    #  Function checks \c robots.txt and respect them.
    #
    #  \param self The object instance itself
    #  \param url  URL to open
    #
    #  \return A tuple of connection object and error message
    #  \see \ref getRemotePage()
    def getConnection(self, url):
        retry = 1
        errormsg = None
        conn = None

        # check robots.txt
        parts = urlparse.urlparse(url)
        if parts.netloc not in self._robots_cache:
            rp = robotparser.RobotFileParser()
            robots_url = '%s://%s/robots.txt' % (
                parts.scheme,
                parts.netloc,
                )
            rp.set_url(robots_url)
            rp.read()
            self._robots_cache[parts.netloc] = rp
        if not self._robots_cache[parts.netloc].can_fetch("*", url):
            errormsg = \
                _(u'Zugriff auf "%s" ist durch robots.txt gesperrt.' % url)
            self.logger.warning(errormsg)
            return (None, errormsg)

        while not (errormsg or conn):

            request = urllib2.Request(url)
            if self.user_agent:
                request.add_header('User-agent', self.user_agent)
            try:
                conn = urllib2.urlopen(request)
            except urllib2.HTTPError, e:
                errormsg = _(u'Es trat der Fehler (%(code)d - %(msg)s) beim Abholen der Daten von %(url)s auf.' % {
                    'code': e.code,
                    'msg':  e.msg,
                    'url':  e.url,
                    })
            except urllib2.URLError, e:
                # check for wrapped socket errors
                if isinstance(e.reason, socket.gaierror):
                    code = e.reason[0]
                    #msg = e.reason[1]
                    if code == socket.EAI_AGAIN and retry <= self.max_retries:
                        self.logger.info(
                            _(u'Es trat ein temporrer Fehler beim Abholen der Daten auf. Der Vorgang wird in %d Sekunden wiederholt'),
                            self.wait,
                            )
                        retry += 1
                        time.sleep(self.wait)
                        continue

                errormsg = _(u'Es trat der Fehler (%(msg)s) beim Abholen der Daten von %(url)s auf.' % {
                    'msg':  e,
                    'url':  url,
                    })
            # catch all other exceptions
            except Exception, e:
                errormsg = _(u'Es trat der Fehler (%(details)s) beim Abholen der Daten von %(url)s auf.' % {
                    'details': e,
                    'url':     url,
                    })

        if errormsg:
            return (None, errormsg)
        else:
            return (conn, None)

    ## Search for a text pattern in a BeautifulSoup.BeautifulSoup instance
    #
    #  \param self The object instance itself
    #  \param soup A BeautifulSoup.BeautifulSoup instance
    #  \param text Pattern to search for
    #
    #  \return The next string element follow to the text pattern or None
    def getStringFromHtml(self, soup, text):
        match = soup.find(text=text)
        if not match:
            return None
        element = match.next
        while not element.string:
            element = element.next

        return element.string.decode().strip()

    ## Convert a timestamp string into a datetime object
    #
    #  Unicode strings for format and date / time information are
    #  converted into plain ascii automatically.
    #
    #  \param self      The object instance itself
    #  \param share     Dictionary with the share information
    #  \param timestamp Format of the date string
    #  \param format    Format of the timestamp string
    #  \param loc       Locale used for decode timestamps, default 'C'
    #
    #  On success \c share['success'] is set to \c True and \c
    #  share['datetime'] contains the datetime object.
    #
    #  On failure \c share['success'] will set to \c False and a detailed
    #  error message will stored in \c share['errormsg']. \c share['datetime']
    #  isn't defined.
    #
    #  \note
    #  A \c ValueError exception is rasing during conversion of the
    #  am / pm suffix because some locale don't know those suffixes.
    #  We switch to 'C' locale if the decoding directive for am / pm
    #  \%p is found in the format string.
    def timestamp2datetime(self, share, timestamp, format, loc=None):
        locale_orig = None
        locale_new = loc
        enc = 'ascii'

        # select C locale to convert "%p" (am/pm) correct
        if not locale_new and '%p' in format:
            locale_new = 'C'

        # save current locale settings and set locale
        locale_orig = locale.setlocale(locale.LC_TIME)

        if locale_new:
            # do nothing for equal locale
            if locale_new == locale_orig:
                locale_orig = None

            else:
                # switch locale settings
                locale.setlocale(locale.LC_TIME, (locale_new, None))

                # adjust "target" encoding
                if locale_new.startswith('de'):
                    enc = 'latin-1'

        else:
            locale_orig = None

        # convert unicode into plain ascii (or latin-1)
        if type(timestamp) == types.UnicodeType:
            timestamp = timestamp.encode(enc, 'replace')
        if type(format) == types.UnicodeType:
            format = format.encode(enc, 'replace')

        try:
            moment = datetime.datetime.strptime(timestamp, format)
        except ValueError, inst:
            # Log details about the failed conversion
            msg = _(
                u'Ungltiges Zeitformat fr Wertpapier "%(name)s" (Format: "%(format)s", Zeitmarke: "%(time)s")' % {
                'format': format,
                'name':   self._getName(share),
                'time':   timestamp,
                })
            self.logger.error(
                _(u'Konvertierung des Datums schlug fehl: %s!'),
                msg
                )
            self.logger.error(
                _(u'Exception Message: %s'),
                inst.message
                )
            self.logger.error(
                _(u'Lokalisierung:     %s'),
                locale.setlocale(locale.LC_TIME).decode('ascii', 'replace')
                )
            share['success']  = False
            share['errormsg'] = msg
        else:
            today = datetime.date.today()

            # check for missing date (year, month and day)
            if moment.date() == datetime.date(1900, 1, 1):
                moment = moment.replace(today.year, today.month, today.day)

            # check for missing year
            elif moment.year == 1900:
                moment = moment.replace(today.year)

            # check for future date
            if moment.date() > today:
                self.logger.warning(
                    _(u'Zuknftiges Datum %(date)s fr Wertpapier "%(name)s" korrigiert.'),
                    {
                        'date': moment.date(),
                        'name': self._getName(share),
                    })
                moment.replace(year=today.year-1)

            share['quote.datetime'] = moment
            share['success'] = True


        # restore original settings
        if locale_orig:
            locale.setlocale(locale.LC_TIME, locale_orig)

    ## Convert a unicode string into a decimal value
    #
    #  \param self   The object instance itself
    #  \param value  Unicode string to convert into decimal value
    #  \param dsep   Decimal separator
    #  \param tsep   Thousands separator
    #
    #  \return Decimal valued
    #
    #  Leading and tailing spaces will be removed automatically. 
    #
    #  The percent sign is cut off and the value is divided by 100.
    #  100% is Decimal("1").
    #
    #  \note \c None will not processed
    #
    #  \b Examples:
    #  \code
    #  >>> self.toDecimal(u'12%')
    #  Decimal("0.12")
    #  >>> self.toDecimal(u'12 %')
    #  Decimal("0.12")
    #  >>> self.toDecimal(u'12,30')
    #  Decimal("12.30")
    #  >>> self.toDecimal(u'12.30')
    #  Decimal("12.30")
    #  >>> self.toDecimal(u'12.000', tsep=u'.')
    #  Decimal("12000")
    #  >>> self.toDecimal(u'12.000%', tsep=u'.')
    #  Decimal("120")
    #  \endcode
    #
    #  \exception ValueError  Raised for not convertible values
    #  \exception TypeError   Raised for non unicode \c value
    def toDecimal(self, value, dsep=u',', tsep=u''):
        # return value if is already a Decimal class
        if isinstance(value, decimal.Decimal):
            return value
                    
        # Convert unicode only
        if type(value) != types.UnicodeType:
            raise TypeError(_(u'Falscher Datentyp %s, erwartet %s' % (type(value), types.UnicodeType)))
                    
        # initialise internal variables
        isPercent = False
        _value = value.strip()
            
        # delete thousand separator
        if tsep:
            _value = _value.replace(tsep, '')
            
        # convert decimal separator to '.'
        if dsep != u'.':
            _value = _value.replace(dsep, u'.')
        
        # check percent sign
        if _value.endswith(u'%'):
            _value = _value[:-1].rstrip()
            isPercent = True
            
        # convert to decimal values
        try:
            _value = decimal.Decimal(_value)
        except decimal.InvalidOperation:
            raise ValueError(
                _(u'Die Zeichenkette "%s" konnte nicht in einem Dezimalwert umgewandelt werden' % (_value))
                )
        except:
            logging.exception(
                _(u'Es trat ein unerwarteter Fehler whrend des Umwandelns von "%s" in einen Dezimalwert auf'),
                _value,
                )
                
        # handle percent
        if isPercent:
            _value = _value / 100

        return _value

    ## Converts an attribute into a decimal value.
    #
    #  Errors will be catched smoothly. \c struck[attr] is deleted in case
    #  of errors.
    #
    #  \param self    The object instance itself
    #  \param struct  Share dict
    #  \param attr    Attribute to store decimal value
    #  \param value   Unicode string or \c BeautifulSoup.NavigableString,
    #                 \c BeautifulSoup.Tag to convert into decimal value
    def attr2decimal(self, struct, attr, value):
        _value = None
        if attr in struct:
            del struct[attr]

        # special handling for some often used objects
        if isinstance(value, BeautifulSoup.Tag):
            value = value.text
        elif isinstance(value, BeautifulSoup.NavigableString):
            value = unicode(value)
        elif type(value) != types.UnicodeType:
            self.logger.warn(
                _(u'Das Attribut %s hat den falschen Typ %s, erwartet wird %s'),
                attr,
                type(value),
                types.UnicodeType,
                )
            return

        try:
            _value = self.toDecimal(value, self.dsep, self.tsep)
        except TypeError:
            self.logger.warn(
                _(u'Das Attribut %s hat den falschen Typ %s, erwartet wird %s'),
                attr,
                type(value),
                types.UnicodeType,
                )
            return
        except ValueError:
            self.logger.warn(
                _(u'Das Attribut %s mit dem Wert %s konnte nicht in einen Dezimalwert umgewandelt werden.'),
                attr,
                value,
                )
            return
        except:
            logging.exception(
                _(u'Es trat ein unerwarteter Fehler whrend des Umwandeln des Attributes %s mit dem Wert %s in einen Dezimalwert auf'),
                attr,
                value,
                )
            return

        # set converted value
        struct[attr] = _value

